/**
 *
 * Copyright (c) 2001, Frank Bu
 *
 * project: LSystem
 * modul:   Test
 * version: $Revision: 1.2 $ $Name:  $
 *
 * CppUnit Testklasse fr Expression.
 */

// disable "identifier was truncated to 255 characters in the debug information"
#pragma warning(disable : 4786)

#include "test.h"
#include "TestParser.h"

#include <string>

#include <Parser.h>
#include <iostream>
#include <math.h>
#include <time.h>
#include <float.h>

using namespace std;


void TestParser::testParse()
{
	Parser parser("1+23*456^3");
	assert(parser.getPostfix() == "1 23 456 3 ^ * +");
	parser.setExpression("(1+23)*456");
	assert(parser.getPostfix() == "1 23 + 456 *");
	parser.setExpression("1+2");
	assert(parser.getPostfix() == "1 2 +");
	parser.setExpression("(1+2*3+4*(3+4))+6");
	assert(parser.getPostfix() == "1 2 3 * + 4 3 4 + * + 6 +");
	parser.setExpression("(1+2*3-4*(3+4))/6");
	assert(parser.getPostfix() == "1 2 3 * + 4 3 4 + * - 6 /");
	parser.setExpression("1.2e-23");
	assert(parser.getPostfix() == "1.2e-23");
	parser.setExpression("1.2e-23*3");
	assert(parser.getPostfix() == "1.2e-23 3 *");
	parser.setExpression("1+1.2e-23");
	assert(parser.getPostfix() == "1 1.2e-23 +");
	parser.setExpression("1+134.2e-23");
	assert(parser.getPostfix() == "1 134.2e-23 +");
	parser.setExpression("1+134.2");
	assert(parser.getPostfix() == "1 134.2 +");
	parser.setExpression("134.2/2");
	assert(parser.getPostfix() == "134.2 2 /");
	parser.setExpression(".23");
	assert(parser.getPostfix() == ".23");
	parser.setExpression(".23-12");
	assert(parser.getPostfix() == ".23 12 -");
	parser.setExpression("12-.23");
	assert(parser.getPostfix() == "12 .23 -");
	parser.setExpression("12-.23E+10");
	assert(parser.getPostfix() == "12 .23E+10 -");
	parser.setExpression("12-.23E10");
	assert(parser.getPostfix() == "12 .23E10 -");
	parser.setExpression("12-.23E10*9");
	assert(parser.getPostfix() == "12 .23E10 9 * -");
	parser.setExpression("12-.23E10+test*9");
	assert(parser.getPostfix() == "12 .23E10 - test 9 * +");
	parser.setExpression("-34");
	assert(parser.getPostfix() == "34 neg");
	parser.setExpression("14*-2^3");
	assert(parser.getPostfix() == "14 2 3 ^ neg *");
}


void TestParser::testEval()
{
	Parser parser("2^3+23*456");
	assertDoublesEqual(pow(2,3)+23*456, parser.eval(), 1e-9);
	parser.setExpression("(1+23)*-456");
	assertDoublesEqual((1+23)*-456, parser.eval(), 1e-9);
	parser.setExpression("-(1+2)^2");
	assertDoublesEqual(-pow(1+2,2), parser.eval(), 1e-9);
	parser.setExpression("(1+2*3+4*(-3+4))-6^3");
	assertDoublesEqual((1+2*3+4*(-3+4))-pow(6,3), parser.eval(), 1e-9);
	parser.setExpression("(1+2*3-4*(3+4))/6");
	assertDoublesEqual((1.0+2.0*3.0-4.0*(3.0+4.0))/6.0, parser.eval(), 1e-9);
}


void TestParser::testVariable()
{
	Parser parser("1+23*test+3^exp");
	parser.setVariable("test", 12.34);
	parser.setVariable("exp", -1.5);
	assertDoublesEqual(1+23*12.34+pow(3,-1.5), parser.eval(), 1e-9);
}


void TestParser::testFunction()
{
	Parser parser("time()");
	assert(parser.eval() > 900000000l && parser.eval() <= time(NULL));
	parser.setExpression("1+sqrt(81)*456");
	assert(parser.eval() == (1+sqrt(81)*456));
	parser.setExpression("atan2(2,3)");
	assertDoublesEqual(atan2(2,3), parser.eval(), 1e-9);
}


void TestParser::testNested()
{
	double a = 10;
	double b = -34.4;
	Parser parser("sqrt(1-2*(a+(b*(3+b-atan2(a,b*sqrt(81))*4e-21/a)+(1/.2e12+43.9*b))))");
	parser.setVariable("a", a);
	parser.setVariable("b", b);
	assertDoublesEqual(sqrt(1-2*(a+(b*(3+b-atan2(a,b*sqrt(81))*4e-21/a)+(1/.2e12+43.9*b)))), parser.eval(), 1e-9);
}


void TestParser::testLogic()
{
	Parser parser("1>2");
	assert(0 == parser.eval());
	parser.setExpression("1<2");
	assert(1 == parser.eval());
	parser.setExpression("1<2-5");
	assert(0 == parser.eval());
	parser.setExpression("1<=1");
	assert(1 == parser.eval());
	parser.setExpression("1<=3");
	assert(1 == parser.eval());
	parser.setExpression("1<=0");
	assert(0 == parser.eval());
	parser.setExpression("1>=0");
	assert(1 == parser.eval());
	parser.setExpression("1>=3");
	assert(0 == parser.eval());
	parser.setExpression("1+2>=3");
	assert(1 == parser.eval());
	parser.setExpression("1=3");
	assert(0 == parser.eval());
	parser.setExpression("3=3");
	assert(1 == parser.eval());
	parser.setExpression("3!=3");
	assert(0 == parser.eval());
	parser.setExpression("2!=3");
	assert(1 == parser.eval());
	parser.setExpression("4<3+2=1");
	assert(1 == parser.eval());
	parser.setExpression("4<3+2!=1");
	assert(0 == parser.eval());
	parser.setExpression("0&0");
	assert(0 == parser.eval());
	parser.setExpression("1&0");
	assert(0 == parser.eval());
	parser.setExpression("1&1");
	assert(1 == parser.eval());
	parser.setExpression("0|0");
	assert(0 == parser.eval());
	parser.setExpression("1|0");
	assert(1 == parser.eval());
	parser.setExpression("1|1");
	assert(1 == parser.eval());
	parser.setExpression("1<2&1<0|1<2");
	assert(1 == parser.eval());
	parser.setExpression("1<2&1<0");
	assert(0 == parser.eval());
	parser.setExpression("1<2&1<0+9");
	assert(1 == parser.eval());
	parser.setExpression("0=2|1=1");
	assert(1 == parser.eval());
	parser.setExpression("0=(2|1)=1");
	assert(0 == parser.eval());
	parser.setExpression("!1");
	assert(0 == parser.eval());
	parser.setExpression("!(1=0)");
	assert(1 == parser.eval());
	parser.setExpression("!0|0");
	assert(1 == parser.eval());
	parser.setExpression("!(0|1)");
	assert(0 == parser.eval());
}


void TestParser::testParseError()
{
	Parser parser("");
	try {
		parser.setExpression("-123. + 3");
		assert(false);
	} catch (SyntaxError& e) {
		assert(strcmp(e.what(), "Expected digit after '.', number: 123.") == 0);
	}
	try {
		parser.setExpression("2 @ 3");
		assert(false);
	} catch (SyntaxError& e) {
		assert(strcmp(e.what(), "Invalid character: @") == 0);
	}
	try {
		parser.setExpression("(1+2*3");
		assert(false);
	} catch (SyntaxError& e) {
		assert(strcmp(e.what(), "Missing ')'.") == 0);
	}
	try {
		parser.setExpression("1,2");
		assert(false);
	} catch (SyntaxError& e) {
		assert(strcmp(e.what(), "',' is allowed within functions only.") == 0);
	}
	try {
		parser.setExpression("7+2)*3");
		assert(false);
	} catch (SyntaxError& e) {
		assert(strcmp(e.what(), "')' found but there is no matching '('.") == 0);
	}
	try {
		parser.setExpression("atan2(1,2,3)");
		assert(false);
	} catch (TooManyArgumentsError& e) {
		assert(strcmp(e.what(), "Too many arguments for function: atan2") == 0);
	}
	try {
		parser.setExpression("1**2");
		assert(false);
	} catch (SyntaxError& e) {
		assert(strcmp(e.what(), "Expecting a variable, function, '(', number, not or negate operator.") == 0);
	}
	try {
		parser.setExpression("1 2");
		assert(false);
	} catch (SyntaxError& e) {
		assert(strcmp(e.what(), "One after another number is not allowed.") == 0);
	}
	try {
		parser.setExpression("atan2(1,)");
		assert(false);
	} catch (SyntaxError& e) {
		assert(strcmp(e.what(), "Expecting a variable, function, '(', number, not or negate operator.") == 0);
	}
	try {
		parser.setExpression("1*2+(*)");
		assert(false);
	} catch (SyntaxError& e) {
		assert(strcmp(e.what(), "Expecting a variable, function, '(', ')', number, not or negate operator.") == 0);
	}
	try {
		parser.setExpression("a b");
		assert(false);
	} catch (SyntaxError& e) {
		assert(strcmp(e.what(), "One after another number is not allowed.") == 0);
	}
	try {
		parser.setExpression("(a+2) b");
		assert(false);
	} catch (SyntaxError& e) {
		assert(strcmp(e.what(), "Expected ')' or operator after ')'.") == 0);
	}
}


void TestParser::testEvalError()
{
	Parser parser("2*a");
	try {
		parser.eval();
		assert(false);
	} catch (VariableNotFound& e) {
		assert(e.getVariableName() == "a");
	}
	parser.setExpression("1+thatsNoFunction(123,456)");
	try {
		parser.eval();
		assert(false);
	} catch (FunctionNotFound& e) {
		assert(e.getFunctionName() == "thatsNoFunction");
	}
	try {
		parser.setExpression("1/0");
		parser.eval();
		assert(false);
	} catch (MathError&)
	{
	}
	try {
		parser.setExpression("sqrt(-2)");
		parser.eval();
		assert(false);
	} catch (MathError&)
	{
	}
	try {
		parser.setExpression("asin(10)");
		parser.eval();
		assert(false);
	} catch (MathError&)
	{
	}
	try {
		parser.setExpression("log(-2)");
		parser.eval();
		assert(false);
	} catch (MathError&)
	{
	}
}


Test *TestParser::suite ()
{
	TestSuite *testSuite = new TestSuite("TestExpression");

    testSuite->addTest(new TestCaller <TestParser> ("testParse", testParse));
    testSuite->addTest(new TestCaller <TestParser> ("testEval", testEval));
    testSuite->addTest(new TestCaller <TestParser> ("testVariable", testVariable));
    testSuite->addTest(new TestCaller <TestParser> ("testFunction", testFunction));
    testSuite->addTest(new TestCaller <TestParser> ("testNested", testNested));
    testSuite->addTest(new TestCaller <TestParser> ("testLogic", testLogic));
    testSuite->addTest(new TestCaller <TestParser> ("testParseError", testParseError));
    testSuite->addTest(new TestCaller <TestParser> ("testEvalError", testEvalError));

	return testSuite;
}
